@@ -6,16 +6,15 @@
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.*
+ * License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.LGPL included in the package of this file.
+ * You can also redistribute it and/or modify it under the terms of 
+ * the GNU General Public License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.GPL included in the package of this file.
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
- * GNU Lesser General Public License for more details.*
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
- * Boston, MA  02111-1307  USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
 
 #include <assert.h>
@@ -216,10 +215,13 @@ void ScimBridgeAgentIMContext::static_finalize ()
     panel_listener = NULL;
 
     shared_imengine = NULL;
+
+    fallback_imengine.reset ();
     fallback_imengine = NULL;
 
+    fallback_imengine_factory.reset ();
     fallback_imengine_factory = NULL;
-
+    
     scim_backend = NULL;
 
     for (vector<ScimBridgeAgentIMContextImpl*>::iterator i = imcontexts.begin (); i != imcontexts.end (); ++i) {
@@ -526,7 +528,7 @@ bool ScimBridgeAgentIMContextImpl::filter_key_event (const KeyEvent &key_event)
     } else {
         focus_in ();
     }
-
+    
     if (!is_enabled ()) {
         return false;
     } else if (get_imengine ()->process_key_event (key_event)) {
@@ -646,8 +648,8 @@ void ScimBridgeAgentIMContextImpl::focus_out ()
     ScimBridgeAgentIMContext *focused_imcontext = static_cast<ScimBridgeAgentIMContext*> (get_imengine ()->get_frontend_data ());
 
     get_imengine ()->set_frontend_data (static_cast<ScimBridgeAgentIMContext*> (this));
-    hide_preedit ();
 
+    hide_preedit ();
     get_imengine ()->focus_out ();
     if (imengine_shared) shared_imengine->reset ();
     panel_listener->turn_off ();
