@@ -6,16 +6,15 @@
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.*
+ * License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.LGPL included in the package of this file.
+ * You can also redistribute it and/or modify it under the terms of 
+ * the GNU General Public License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.GPL included in the package of this file.
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
- * GNU Lesser General Public License for more details.*
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
- * Boston, MA  02111-1307  USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
 
 #include <alloca.h>
@@ -99,11 +98,17 @@ static retval_t launch_agent ()
 }
 
 
-/* Message Hundlers */
+/* Message Handlers */
 static retval_t received_message_unknown (const ScimBridgeMessage *message)
 {
     const char *header = scim_bridge_message_get_header (message);
-    scim_bridge_perrorln ("Unknown message: %s", header);
+    scim_bridge_perror ("Unknown message: %s", header);
+    
+    int i;
+    for (i = 0; i < scim_bridge_message_get_argument_count (message); ++i) {
+        scim_bridge_perror (" %s", scim_bridge_message_get_argument (message, i));
+    }
+    scim_bridge_perrorln ("");
 
     // Just ignore it.
     return RETVAL_SUCCEEDED;
@@ -167,6 +172,14 @@ static retval_t received_message_imcontext_deregistered (const ScimBridgeMessage
 
 static retval_t received_message_imcontext_reseted (const ScimBridgeMessage *message)
 {
+    const char *header = scim_bridge_message_get_header (message);
+
+    if (pending_response.status == RESPONSE_PENDING && strcmp (pending_response.header, header) == 0) {
+        pending_response.status = RESPONSE_SUCCEEDED;
+    } else {
+        scim_bridge_perrorln ("The message is recieved in a wrong context: %s", header);
+    }
+    
     return RETVAL_SUCCEEDED;
 }
 
@@ -196,6 +209,14 @@ static retval_t received_message_key_event_handled (const ScimBridgeMessage *mes
 
 static retval_t received_message_focus_changed (const ScimBridgeMessage *message)
 {
+    const char *header = scim_bridge_message_get_header (message);
+
+    if (pending_response.status == RESPONSE_PENDING && strcmp (pending_response.header, header) == 0) {
+        pending_response.status = RESPONSE_SUCCEEDED;
+    } else {
+        scim_bridge_perrorln ("The message is recieved in a wrong context: %s", header);
+    }
+
     return RETVAL_SUCCEEDED;
 }
 
@@ -382,7 +403,7 @@ static retval_t received_message_set_preedit_attributes (const ScimBridgeMessage
                     if (color_str_length == 6) {
                         color_val = 0;
                         int j;
-                        for (j = 6; j >= 0; --j) {
+                        for (j = 0; j < 6; ++j) {
                             color_val <<= 4;
                             switch (color_str[j]) {
                                 case '0':
@@ -447,14 +468,14 @@ static retval_t received_message_set_preedit_attributes (const ScimBridgeMessage
                     }
                 }
 
-                if (color_val >= 0xFFFFFF) {
-                    scim_bridge_perrorln ("Invalid string for a color: %s", attribute_value_str);
+                if (color_val > 0xFFFFFF) {
+                    scim_bridge_perrorln ("An invalid string for a color: %s", attribute_value_str);
                     scim_bridge_attribute_set_type (attribute, ATTRIBUTE_NONE);
                     scim_bridge_attribute_set_value (attribute, SCIM_BRIDGE_ATTRIBUTE_DECORATE_NONE);
                 } else {
-                    const unsigned int red = (color_val && 0xFF0000) >> 16;
-                    const unsigned int green = (color_val && 0x00FF00) >> 8;
-                    const unsigned int blue = (color_val && 0x0000FF) >> 0;
+                    const unsigned int red = (color_val & 0xFF0000) >> 16;
+                    const unsigned int green = (color_val & 0x00FF00) >> 8;
+                    const unsigned int blue = (color_val & 0x0000FF) >> 0;
 
                     scim_bridge_attribute_set_color (attribute, red, green, blue);
                 }
@@ -1399,6 +1420,11 @@ retval_t scim_bridge_client_reset_imcontext (const ScimBridgeClientIMContext *im
         return RETVAL_FAILED;
     }
 
+    if (pending_response.status != RESPONSE_DONE) {
+        scim_bridge_perrorln ("Another command is pending...");
+        return RETVAL_FAILED;
+    }
+
     if (!scim_bridge_client_is_messenger_opened ()) {
         scim_bridge_perrorln ("The messenger is closed");
         return RETVAL_FAILED;
@@ -1422,20 +1448,29 @@ retval_t scim_bridge_client_reset_imcontext (const ScimBridgeClientIMContext *im
         }
     }
 
+    pending_response.header = SCIM_BRIDGE_MESSAGE_IMCONTEXT_RESETED;
+    pending_response.status = RESPONSE_PENDING;
+
     while (pending_response.status == RESPONSE_PENDING) {
         if (scim_bridge_client_read_and_dispatch ()) {
             scim_bridge_perrorln ("An IOException at scim_bridge_client_reset_imcontext ()");
             pending_response.header = NULL;
             pending_response.status = RESPONSE_DONE;
-
             return RETVAL_FAILED;
         }
     }
 
-    scim_bridge_pdebugln (6, "reseted: id = %d", id);
-    pending_response.header = NULL;
-    pending_response.status = RESPONSE_DONE;
-    return RETVAL_SUCCEEDED;
+    if (pending_response.status == RESPONSE_SUCCEEDED) {
+        scim_bridge_pdebugln (6, "reseted: id = %d", id);
+        pending_response.header = NULL;
+        pending_response.status = RESPONSE_DONE;
+        return RETVAL_SUCCEEDED;
+    } else {
+        scim_bridge_perrorln ("An unknown error occurred at scim_bridge_client_reset_imcontext ()");
+        pending_response.header = NULL;
+        pending_response.status = RESPONSE_DONE;
+        return RETVAL_FAILED;
+    }
 }
 
 
@@ -1648,6 +1683,11 @@ retval_t scim_bridge_client_change_focus (const ScimBridgeClientIMContext *imcon
         return RETVAL_FAILED;
     }
 
+    if (pending_response.status != RESPONSE_DONE) {
+        scim_bridge_perrorln ("Another command is pending...");
+        return RETVAL_FAILED;
+    }
+
     scim_bridge_pdebugln (5, "Sending 'change_focus' message: ic_id = %d, focus_in = %s", id, focus_in ? "true":"false");
 
     ScimBridgeMessage *message = scim_bridge_alloc_message (SCIM_BRIDGE_MESSAGE_CHANGE_FOCUS, 2);
@@ -1663,6 +1703,10 @@ retval_t scim_bridge_client_change_focus (const ScimBridgeClientIMContext *imcon
     free (ic_id_str);
     free (focus_in_str);
 
+    pending_response.header = SCIM_BRIDGE_MESSAGE_FOCUS_CHANGED;
+    pending_response.consumed = FALSE;
+    pending_response.status = RESPONSE_PENDING;
+    
     scim_bridge_messenger_push_message (messenger, message);
     scim_bridge_free_message (message);
 
@@ -1674,10 +1718,26 @@ retval_t scim_bridge_client_change_focus (const ScimBridgeClientIMContext *imcon
         }
     }
 
-    scim_bridge_pdebugln (6, "the focus changed: id = %d", id);
-    pending_response.header = NULL;
-    pending_response.status = RESPONSE_DONE;
-    return RETVAL_SUCCEEDED;
+    while (pending_response.status == RESPONSE_PENDING) {
+        if (scim_bridge_client_read_and_dispatch ()) {
+            scim_bridge_perrorln ("An IOException at scim_bridge_client_change_focus ()");
+            pending_response.header = NULL;
+            pending_response.status = RESPONSE_DONE;
+            return RETVAL_FAILED;
+        }
+    }
+    
+    if (pending_response.status == RESPONSE_SUCCEEDED) {
+        scim_bridge_pdebugln (6, "The focus changed: id = %d", id);
+        pending_response.header = NULL;
+        pending_response.status = RESPONSE_DONE;
+        return RETVAL_SUCCEEDED;
+    } else {
+        scim_bridge_perrorln ("An unknown error occurred at scim_bridge_client_change_focus ()");
+        pending_response.header = NULL;
+        pending_response.status = RESPONSE_DONE;
+        return RETVAL_FAILED;
+    }
 }
 
 
@@ -1811,7 +1871,7 @@ retval_t scim_bridge_client_set_preedit_mode (const ScimBridgeClientIMContext *i
         pending_response.status = RESPONSE_DONE;
         return RETVAL_FAILED;
     } else {
-        scim_bridge_pdebugln (6, "the preedit mode changed: id = %d", id);
+        scim_bridge_pdebugln (6, "The preedit mode changed: id = %d", id);
         pending_response.header = NULL;
         pending_response.status = RESPONSE_DONE;
         return RETVAL_SUCCEEDED;
