@@ -6,16 +6,15 @@
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.*
+ * License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.LGPL included in the package of this file.
+ * You can also redistribute it and/or modify it under the terms of 
+ * the GNU General Public License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.GPL included in the package of this file.
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
- * GNU Lesser General Public License for more details.*
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
- * Boston, MA  02111-1307  USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
 
 #include <alloca.h>
@@ -53,7 +52,10 @@ struct _ScimBridgeClientIMContext
     size_t preedit_attribute_capacity;
 
     unsigned int preedit_cursor_position;
+    boolean preedit_cursor_flicking;
+    
     boolean preedit_shown;
+    boolean preedit_started;
 
     char *commit_string;
     size_t commit_string_capacity;
@@ -66,6 +68,7 @@ struct _ScimBridgeClientIMContext
     int window_y;
 };
 
+
 /* Private variables */
 static GdkColor preedit_normal_background;
 static GdkColor preedit_normal_foreground;
@@ -76,14 +79,14 @@ static GType class_type = 0;
 static GObjectClass *root_klass = NULL;
 
 static ScimBridgeClientIMContext *focused_imcontext = NULL;
+static GtkWidget *focused_widget = NULL;
 
 static GtkIMContext *fallback_imcontext = NULL;
-static gulong  fallback_commit_handler;
+static gulong fallback_commit_handler;
 
 static guint key_snooper_id = 0;
 static boolean key_snooper_used = FALSE;
 
-
 /* Class functions */
 static void scim_bridge_client_imcontext_class_initialize (ScimBridgeClientIMContextClass *klass, gpointer *klass_data);
 
@@ -114,7 +117,8 @@ static retval_t filter_key_event (ScimBridgeClientIMContext *imcontext, GdkEvent
     scim_bridge_pdebugln (5, "filter_key_event ()");
     
     if (focused_imcontext != imcontext) scim_bridge_client_imcontext_focus_in (GTK_IM_CONTEXT (imcontext));
-
+    focused_widget = gtk_get_event_widget ((GdkEvent*) event);
+    
     if (scim_bridge_client_is_messenger_opened ()) {
         ScimBridgeKeyEvent *bridge_key_event = scim_bridge_alloc_key_event ();
         scim_bridge_key_event_gdk_to_bridge (bridge_key_event, imcontext->client_window, event);
@@ -136,7 +140,7 @@ static retval_t filter_key_event (ScimBridgeClientIMContext *imcontext, GdkEvent
 
 static retval_t set_cursor_location (ScimBridgeClientIMContext *imcontext, int window_x, int window_y, int cursor_x, int cursor_y)
 {
-    scim_bridge_pdebugln (5, "update_cursor_location ()");
+    scim_bridge_pdebugln (5, "set_cursor_location ()");
 
     if (imcontext->window_x == window_x && imcontext->window_y == window_y && imcontext->cursor_x == cursor_x && imcontext->cursor_y == cursor_y) {
         return RETVAL_SUCCEEDED;
@@ -166,7 +170,7 @@ static gboolean key_snooper (GtkWidget *widget, GdkEventKey *event, gpointer dat
 {
     scim_bridge_pdebugln (7, "key_snooper ()");
 
-    if (scim_bridge_client_is_messenger_opened () && focused_imcontext != NULL) {
+    if (!event->send_event && scim_bridge_client_is_messenger_opened () && focused_imcontext != NULL) {
         if (focused_imcontext->client_window != NULL) {
             int new_window_x;
             int new_window_y;
@@ -207,6 +211,19 @@ static boolean is_key_snooper_enabled ()
     return key_snooper_enabled;
 }
 
+static boolean is_precise_cursor_enabled ()
+{
+    static boolean first_time = TRUE;
+    static boolean precise_cursor_enabled = TRUE;
+
+    if (first_time) {
+        char *env_precise_cursor_enabled = getenv ("SCIM_BRIDGE_PRECISE_CURSOR_ENABLED");
+        if (env_precise_cursor_enabled != NULL) scim_bridge_string_to_boolean (&precise_cursor_enabled, env_precise_cursor_enabled);
+        first_time = FALSE;
+    }
+
+    return precise_cursor_enabled;
+}
 
 /* Bindings */
 void scim_bridge_client_imcontext_set_id (ScimBridgeClientIMContext *imcontext, scim_bridge_imcontext_id_t new_id)
@@ -223,6 +240,9 @@ scim_bridge_imcontext_id_t scim_bridge_client_imcontext_get_id (const ScimBridge
 
 void scim_bridge_client_imcontext_set_preedit_string (ScimBridgeClientIMContext *imcontext, const char *preedit_string)
 {
+    if (imcontext->preedit_string != NULL && preedit_string != NULL && !strcmp (imcontext->preedit_string, preedit_string))
+        return;
+
     size_t preedit_string_length;
     if (preedit_string != NULL) {
         preedit_string_length = strlen (preedit_string);
@@ -255,7 +275,6 @@ void scim_bridge_client_imcontext_set_preedit_shown (ScimBridgeClientIMContext *
     }
 }
 
-
 void scim_bridge_client_imcontext_set_preedit_cursor_position (ScimBridgeClientIMContext *imcontext, int cursor_position)
 {
     imcontext->preedit_cursor_position = cursor_position;
@@ -284,7 +303,30 @@ void scim_bridge_client_imcontext_set_preedit_attributes (ScimBridgeClientIMCont
 
 void scim_bridge_client_imcontext_update_preedit (ScimBridgeClientIMContext *imcontext)
 {
-    g_signal_emit_by_name ((ScimBridgeClientIMContext*) imcontext, "preedit_changed");
+    if (imcontext->preedit_shown && !imcontext->preedit_started) {
+        g_signal_emit_by_name ((ScimBridgeClientIMContext*) imcontext, "preedit-start");
+        imcontext->preedit_started = TRUE;
+    }
+    
+    if (is_precise_cursor_enabled ()) {
+        const size_t old_cursor_position = imcontext->preedit_cursor_position;
+        imcontext->preedit_cursor_position = 0;
+        if (imcontext->preedit_string != NULL)
+            imcontext->preedit_cursor_position = g_utf8_strlen (imcontext->preedit_string, -1);
+        
+        imcontext->preedit_cursor_flicking = TRUE;
+        g_signal_emit_by_name ((ScimBridgeClientIMContext*) imcontext, "preedit-changed");
+        
+        imcontext->preedit_cursor_position = old_cursor_position;
+        imcontext->preedit_cursor_flicking = FALSE;
+    }
+    
+    g_signal_emit_by_name ((ScimBridgeClientIMContext*) imcontext, "preedit-changed");
+    
+    if (!imcontext->preedit_shown && imcontext->preedit_started) {
+        g_signal_emit_by_name ((ScimBridgeClientIMContext*) imcontext, "preedit-end");
+        imcontext->preedit_started = FALSE;
+    }
 }
 
 
@@ -382,10 +424,23 @@ boolean scim_bridge_client_imcontext_replace_surrounding_text (ScimBridgeClientI
 
 
 void scim_bridge_client_imcontext_forward_key_event (ScimBridgeClientIMContext *imcontext, const ScimBridgeKeyEvent *key_event)
-{
+{ 
     GdkEventKey gdk_event;
     scim_bridge_key_event_bridge_to_gdk (&gdk_event, imcontext->client_window, key_event);
-    gdk_event_put ((GdkEvent*) &gdk_event);
+    gdk_event.send_event = TRUE;
+    if (imcontext == focused_imcontext && focused_widget != NULL) {
+        const char *signal_name = NULL;
+        if (scim_bridge_key_event_is_pressed (key_event)) {
+            signal_name = "key-press-event";
+        } else {
+            signal_name = "key-release-event";
+        }
+        
+        gboolean consumed;
+        g_signal_emit_by_name (focused_widget, signal_name, &gdk_event, &consumed);
+    } else {
+        gdk_event_put ((GdkEvent*) &gdk_event);
+    }
 }
 
 
@@ -494,6 +549,12 @@ void scim_bridge_client_imcontext_initialize (ScimBridgeClientIMContext *imconte
 {
     scim_bridge_pdebugln (5, "scim_bridge_client_imcontext_initialize  ()");
 
+    imcontext->preedit_shown = FALSE;
+    imcontext->preedit_started = FALSE;
+    
+    imcontext->preedit_cursor_position = 0;
+    imcontext->preedit_cursor_flicking = FALSE;
+
     imcontext->preedit_string = malloc (sizeof (char));
     imcontext->preedit_string[0] = '\0';
     imcontext->preedit_string_capacity = 0;
@@ -560,8 +621,8 @@ gboolean scim_bridge_client_imcontext_filter_key_event (GtkIMContext *context, G
     scim_bridge_pdebugln (8, "scim_bridge_client_imcontext_filter_key_event ()");
 
     ScimBridgeClientIMContext *imcontext = SCIM_BRIDGE_CLIENT_IMCONTEXT (context);
-
-    if (scim_bridge_client_is_messenger_opened () && imcontext != NULL && !key_snooper_used) {
+    
+    if (!event->send_event && scim_bridge_client_is_messenger_opened () && imcontext != NULL && !key_snooper_used) {
 
         if (imcontext->client_window != NULL) {
             int new_window_x;
@@ -616,9 +677,10 @@ void scim_bridge_client_imcontext_get_preedit_string (GtkIMContext *context, gch
     scim_bridge_pdebugln (4, "scim_bridge_client_imcontext_get_preedit_string ()");
 
     ScimBridgeClientIMContext *imcontext = SCIM_BRIDGE_CLIENT_IMCONTEXT (context);
-
+    
     if (scim_bridge_client_is_messenger_opened () && imcontext != NULL && imcontext->preedit_shown) {
         const size_t preedit_string_length = strlen (imcontext->preedit_string);
+        const size_t preedit_wstring_length = g_utf8_strlen (imcontext->preedit_string, -1);
         
         if (str) {
             if (preedit_string_length > 0) {
@@ -627,40 +689,48 @@ void scim_bridge_client_imcontext_get_preedit_string (GtkIMContext *context, gch
                 *str = g_strdup ("");
             }
         }
+        
         if (cursor_pos) {
-            const size_t preedit_length = g_utf8_strlen (imcontext->preedit_string, -1);
-            if (imcontext->preedit_cursor_position > preedit_length) {
-                *cursor_pos = preedit_length;
+            if (imcontext->preedit_cursor_position > preedit_wstring_length) {
+                *cursor_pos = preedit_wstring_length;
             } else {
                 *cursor_pos = imcontext->preedit_cursor_position;
             }
         }
-
+        
         if (pango_attrs) {
             *pango_attrs = pango_attr_list_new ();
-
-            gboolean underline_exists = FALSE;
-
+            
+            boolean *has_attribute = alloca (sizeof (boolean) * preedit_string_length);
             int i;
+            for (i = 0; i < preedit_string_length; ++i) {
+                has_attribute[i] = FALSE;
+            }
+
             for (i = 0; i < imcontext->preedit_attribute_count; ++i) {
                 const ScimBridgeAttribute *attr = imcontext->preedit_attributes[i];
                 const int begin_pos = scim_bridge_attribute_get_begin (attr);
                 const int end_pos = scim_bridge_attribute_get_end (attr);
 
-                const int start_index = g_utf8_offset_to_pointer (imcontext->preedit_string, begin_pos) - imcontext->preedit_string;
-                const int end_index = g_utf8_offset_to_pointer (imcontext->preedit_string, end_pos) - imcontext->preedit_string;
+                if (begin_pos <= end_pos && 0 <= begin_pos && end_pos <= preedit_wstring_length) {
+                    const int start_index = g_utf8_offset_to_pointer (imcontext->preedit_string, begin_pos) - imcontext->preedit_string;
+                    const int end_index = g_utf8_offset_to_pointer (imcontext->preedit_string, end_pos) - imcontext->preedit_string;
 
-                if (start_index <= end_index && 0 <= start_index && end_index <= preedit_string_length) {
                     const scim_bridge_attribute_type_t attr_type = scim_bridge_attribute_get_type (attr);
                     const scim_bridge_attribute_value_t attr_value = scim_bridge_attribute_get_value (attr);
+
+                    boolean valid_attribute = FALSE;
                     if (attr_type == ATTRIBUTE_DECORATE) {
                         if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_UNDERLINE) {
+                            valid_attribute = TRUE;
+
                             PangoAttribute *pango_attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
                             pango_attr->start_index = start_index;
                             pango_attr->end_index = end_index;
                             pango_attr_list_insert (*pango_attrs, pango_attr);
-                            underline_exists = TRUE;
                         } else if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_REVERSE) {
+                            valid_attribute = TRUE;
+
                             PangoAttribute *pango_attr0 = pango_attr_foreground_new (preedit_normal_background.red, preedit_normal_background.green, preedit_normal_background.blue);
                             pango_attr0->start_index = start_index;
                             pango_attr0->end_index = end_index;
@@ -671,6 +741,8 @@ void scim_bridge_client_imcontext_get_preedit_string (GtkIMContext *context, gch
                             pango_attr1->end_index = end_index;
                             pango_attr_list_insert (*pango_attrs, pango_attr1);
                         } else if (attr_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_HIGHLIGHT) {
+                            valid_attribute = TRUE;
+
                             PangoAttribute *pango_attr0 = pango_attr_foreground_new (preedit_active_foreground.red, preedit_active_foreground.green, preedit_active_foreground.blue);
                             pango_attr0->start_index = start_index;
                             pango_attr0->end_index = end_index;
@@ -684,6 +756,8 @@ void scim_bridge_client_imcontext_get_preedit_string (GtkIMContext *context, gch
                             scim_bridge_perrorln ("Unknown preedit decoration!");
                         }
                     } else if (attr_type == ATTRIBUTE_FOREGROUND) {
+                        valid_attribute = TRUE;
+
                         const unsigned int red = scim_bridge_attribute_get_red (attr) * 256;
                         const unsigned int green = scim_bridge_attribute_get_green (attr) * 256;
                         const unsigned int blue = scim_bridge_attribute_get_blue (attr) * 256;
@@ -693,6 +767,8 @@ void scim_bridge_client_imcontext_get_preedit_string (GtkIMContext *context, gch
                         pango_attr->end_index = end_index;
                         pango_attr_list_insert (*pango_attrs, pango_attr);
                     } else if (attr_type == ATTRIBUTE_BACKGROUND) {
+                        valid_attribute = TRUE;
+
                         const unsigned int red = scim_bridge_attribute_get_red (attr) * 256;
                         const unsigned int green = scim_bridge_attribute_get_green (attr) * 256;
                         const unsigned int blue = scim_bridge_attribute_get_blue (attr) * 256;
@@ -702,15 +778,26 @@ void scim_bridge_client_imcontext_get_preedit_string (GtkIMContext *context, gch
                         pango_attr->end_index = end_index;
                         pango_attr_list_insert (*pango_attrs, pango_attr);
                     }
+
+                    if (valid_attribute) {
+                        int j;
+                        for (j = start_index; j < end_index; ++j) {
+                            has_attribute[j] = TRUE;
+                        }
+                    }
                 }
 
             }
-            /* If there is no underline at all, then draw underline under the whole preedit string.*/
-            if (!underline_exists) {
-                PangoAttribute *pango_attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
-                pango_attr->start_index = 0;
-                pango_attr->end_index = preedit_string_length;
-                pango_attr_list_insert (*pango_attrs, pango_attr);
+
+            // Add underlines for the all characters without attributes.
+            for (i = 0; i < preedit_string_length; ++i) {
+                if (has_attribute[i] == FALSE) {
+                    PangoAttribute *pango_attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
+                    pango_attr->start_index = i;
+                    for (; i < preedit_string_length && has_attribute[i] == FALSE; ++i);
+                    pango_attr->end_index = i;
+                    pango_attr_list_insert (*pango_attrs, pango_attr);
+                }
             }
         }
     } else {
@@ -726,9 +813,8 @@ void scim_bridge_client_imcontext_focus_in (GtkIMContext *context)
     scim_bridge_pdebugln (8, "scim_bridge_client_imcontext_focus_in ()");
 
     ScimBridgeClientIMContext *imcontext = SCIM_BRIDGE_CLIENT_IMCONTEXT (context);
-
+    
     if (focused_imcontext != NULL) scim_bridge_client_imcontext_focus_out (GTK_IM_CONTEXT (focused_imcontext));
-
     focused_imcontext = imcontext;
 
     if (!scim_bridge_client_is_messenger_opened () && scim_bridge_client_is_reconnection_enabled ()) {
@@ -753,6 +839,7 @@ void scim_bridge_client_imcontext_focus_out (GtkIMContext *context)
     scim_bridge_pdebugln (8, "scim_bridge_client_imcontext_focus_out ()");
 
     ScimBridgeClientIMContext *imcontext = SCIM_BRIDGE_CLIENT_IMCONTEXT (context);
+    focused_widget = NULL;
 
     focused_imcontext = imcontext;
     if (imcontext->preedit_shown) {
@@ -797,6 +884,8 @@ void scim_bridge_client_imcontext_set_cursor_location (GtkIMContext *context, Gd
     scim_bridge_pdebugln (4, "scim_bridge_client_imcontext_set_cursor_location ()");
 
     ScimBridgeClientIMContext *imcontext = SCIM_BRIDGE_CLIENT_IMCONTEXT (context);
+    if (imcontext->preedit_cursor_flicking)
+        return;
 
     if (imcontext != NULL && imcontext->client_window != NULL) {
         const int new_cursor_x = area->x + area->width;
