@@ -6,16 +6,15 @@
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.*
+ * License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.LGPL included in the package of this file.
+ * You can also redistribute it and/or modify it under the terms of 
+ * the GNU General Public License as published by the Free Software Foundation and 
+ * appearing in the file LICENSE.GPL included in the package of this file.
+ *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
- * GNU Lesser General Public License for more details.*
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
- * Boston, MA  02111-1307  USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
 
 #include "scim-bridge-output.h"
@@ -32,6 +31,8 @@ class ScimBridgeClientIMContextImpl;
 
 static ScimBridgeClientIMContextImpl *focused_imcontext = NULL;
 
+static bool key_event_forwarded = false;
+
 
 /* Class Definition */
 class ScimBridgeClientIMContextImpl: public _ScimBridgeClientIMContext
@@ -42,9 +43,7 @@ class ScimBridgeClientIMContextImpl: public _ScimBridgeClientIMContext
         ScimBridgeClientIMContextImpl ();
         ~ScimBridgeClientIMContextImpl ();
 
-#if defined(Q_WS_X11)
         bool x11FilterEvent (QWidget *widget, XEvent *event);
-#endif
         bool filterEvent (const QEvent *event);
         void mouseHandler (int offset, QMouseEvent *event);
 
@@ -181,25 +180,51 @@ void ScimBridgeClientIMContextImpl::mouseHandler (int offset, QMouseEvent *event
 }
 
 
-#if defined(Q_WS_X11)
 bool ScimBridgeClientIMContextImpl::x11FilterEvent (QWidget *widget, XEvent *event)
 {
+    scim_bridge_pdebugln (5, "ScimBridgeClientIMContextImpl::x11FilterEvent ()");
+
+#ifdef Q_WS_X11
+    if (key_event_forwarded || (event->type != XKeyPress && event->type != XKeyRelease)) return false;
+    
+    if (focused_imcontext != this) focus_in ();
+
+    if (scim_bridge_client_is_messenger_opened ()) {
+        ScimBridgeKeyEvent *bridge_key_event = scim_bridge_key_event_x11_to_bridge (event);
+
+        boolean consumed = FALSE;
+        const retval_t retval_error = scim_bridge_client_handle_key_event (this, bridge_key_event, &consumed);
+
+        scim_bridge_free_key_event (bridge_key_event);
+
+        if (retval_error) {
+            scim_bridge_perrorln ("An IOException at x11FilterEvent ()");
+        } else {
+            return consumed;
+        }
+    }
+#endif
+
     return false;
 }
-#endif
 
 
 bool ScimBridgeClientIMContextImpl::filterEvent (const QEvent *event)
 {
     scim_bridge_pdebugln (5, "ScimBridgeClientIMContextImpl::filterEvent ()");
 
-    if (event->type() != QEvent::KeyPress && event->type() != QEvent::KeyRelease) return false;
+#ifndef Q_WS_X11
+    if (key_event_forwarded || (event->type () != QEvent::KeyPress && event->type () != QEvent::KeyRelease)) return false;
     
     if (focused_imcontext != this) focus_in ();
 
     if (scim_bridge_client_is_messenger_opened ()) {
-        ScimBridgeKeyEvent *bridge_key_event = scim_bridge_key_event_qt_to_bridge ((const QKeyEvent*) event);
-
+        const QKeyEvent *key_event = static_cast<const QKeyEvent*> (event);
+        ScimBridgeKeyEvent *bridge_key_event = scim_bridge_key_event_qt_to_bridge (key_event);
+        
+        if (scim_bridge_key_event_get_code (bridge_key_event) == SCIM_BRIDGE_KEY_CODE_NullKey)
+            return false;
+        
         boolean consumed = FALSE;
         const retval_t retval_error = scim_bridge_client_handle_key_event (this, bridge_key_event, &consumed);
 
@@ -211,8 +236,9 @@ bool ScimBridgeClientIMContextImpl::filterEvent (const QEvent *event)
             return consumed;
         }
     }
+#endif
 
-    return FALSE;
+    return false;
 }
 
 
@@ -323,8 +349,20 @@ void ScimBridgeClientIMContextImpl::forward_key_event (const ScimBridgeKeyEvent
 
     QWidget *focused_widget = qApp->focusWidget ();
     if (focused_widget != NULL) {
-        QKeyEvent *qt_key_event = scim_bridge_key_event_bridge_to_qt (key_event);
-        QApplication::sendEvent (focused_widget, qt_key_event);
+        key_event_forwarded = true;
+#ifdef Q_WS_X11
+        const WId window_id = focused_widget->winId ();
+        XEvent *x_event = scim_bridge_key_event_bridge_to_x11 (key_event, qt_xdisplay (), window_id);
+        qApp->x11ProcessEvent (x_event);
+        free (x_event);
+#else
+        QKeyEvent *forwarded_key_event = scim_bridge_key_event_bridge_to_qt (key_event);
+        QApplication::sendEvent (focused_widget, forwarded_key_event);
+        delete forwarded_key_event;
+#endif
+        key_event_forwarded = false;
+    } else {
+        scim_bridge_pdebugln (4, "No widget is focused");
     }
 }
 
@@ -374,7 +412,6 @@ void ScimBridgeClientIMContextImpl::set_preedit_attributes (ScimBridgeAttribute*
         const scim_bridge_attribute_value_t attribute_value = scim_bridge_attribute_get_value (attribute);
 
         if (attribute_type == ATTRIBUTE_DECORATE && (attribute_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_HIGHLIGHT || attribute_value == SCIM_BRIDGE_ATTRIBUTE_DECORATE_REVERSE)) {
-
             preedit_selected_offset = attribute_begin;
             preedit_selected_length = attribute_end - attribute_begin;
             break;
@@ -392,10 +429,10 @@ void ScimBridgeClientIMContextImpl::update_preedit ()
 
         const size_t preedit_length = preedit_string.length ();
 
-        int cursor_position = preedit_cursor_position;
+        size_t cursor_position = preedit_cursor_position;
         if (cursor_position > preedit_length) cursor_position = preedit_length;
 
-        int selected_length = preedit_selected_length;
+        size_t selected_length = preedit_selected_length;
         if (cursor_position + selected_length > preedit_length) selected_length = preedit_length - cursor_position;
 
         sendIMEvent (QEvent::IMCompose, preedit_string, cursor_position, selected_length);
